# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vega/wallet/v1/wallet.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...commands.v1 import commands_pb2 as vega_dot_commands_dot_v1_dot_commands__pb2
from ...commands.v1 import (
    validator_commands_pb2 as vega_dot_commands_dot_v1_dot_validator__commands__pb2,
)
from ...commands.v1 import data_pb2 as vega_dot_commands_dot_v1_dot_data__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b"\n\x1bvega/wallet/v1/wallet.proto\x12\x0evega.wallet.v1\x1a\x1fvega/commands/v1/commands.proto\x1a)vega/commands/v1/validator_commands.proto\x1a\x1bvega/commands/v1/data.proto\"\xb7\x12\n\x18SubmitTransactionRequest\x12\x17\n\x07pub_key\x18\x01 \x01(\tR\x06pubKey\x12\x1c\n\tpropagate\x18\x02 \x01(\x08R\tpropagate\x12O\n\x10order_submission\x18\xe9\x07 \x01(\x0b\x32!.vega.commands.v1.OrderSubmissionH\x00R\x0forderSubmission\x12U\n\x12order_cancellation\x18\xea\x07 \x01(\x0b\x32#.vega.commands.v1.OrderCancellationH\x00R\x11orderCancellation\x12L\n\x0forder_amendment\x18\xeb\x07 \x01(\x0b\x32 .vega.commands.v1.OrderAmendmentH\x00R\x0eorderAmendment\x12X\n\x13withdraw_submission\x18\xec\x07 \x01(\x0b\x32$.vega.commands.v1.WithdrawSubmissionH\x00R\x12withdrawSubmission\x12X\n\x13proposal_submission\x18\xed\x07 \x01(\x0b\x32$.vega.commands.v1.ProposalSubmissionH\x00R\x12proposalSubmission\x12L\n\x0fvote_submission\x18\xee\x07 \x01(\x0b\x32 .vega.commands.v1.VoteSubmissionH\x00R\x0evoteSubmission\x12w\n\x1eliquidity_provision_submission\x18\xef\x07 \x01(\x0b\x32..vega.commands.v1.LiquidityProvisionSubmissionH\x00R\x1cliquidityProvisionSubmission\x12X\n\x13\x64\x65legate_submission\x18\xf0\x07 \x01(\x0b\x32$.vega.commands.v1.DelegateSubmissionH\x00R\x12\x64\x65legateSubmission\x12^\n\x15undelegate_submission\x18\xf1\x07 \x01(\x0b\x32&.vega.commands.v1.UndelegateSubmissionH\x00R\x14undelegateSubmission\x12}\n liquidity_provision_cancellation\x18\xf2\x07 \x01(\x0b\x32\x30.vega.commands.v1.LiquidityProvisionCancellationH\x00R\x1eliquidityProvisionCancellation\x12t\n\x1dliquidity_provision_amendment\x18\xf3\x07 \x01(\x0b\x32-.vega.commands.v1.LiquidityProvisionAmendmentH\x00R\x1bliquidityProvisionAmendment\x12\x39\n\x08transfer\x18\xf4\x07 \x01(\x0b\x32\x1a.vega.commands.v1.TransferH\x00R\x08transfer\x12L\n\x0f\x63\x61ncel_transfer\x18\xf5\x07 \x01(\x0b\x32 .vega.commands.v1.CancelTransferH\x00R\x0e\x63\x61ncelTransfer\x12\x46\n\rannounce_node\x18\xf6\x07 \x01(\x0b\x32\x1e.vega.commands.v1.AnnounceNodeH\x00R\x0c\x61nnounceNode\x12h\n\x19\x62\x61tch_market_instructions\x18\xf7\x07 \x01(\x0b\x32).vega.commands.v1.BatchMarketInstructionsH\x00R\x17\x62\x61tchMarketInstructions\x12:\n\tnode_vote\x18\xd2\x0f \x01(\x0b\x32\x1a.vega.commands.v1.NodeVoteH\x00R\x08nodeVote\x12I\n\x0enode_signature\x18\xd3\x0f \x01(\x0b\x32\x1f.vega.commands.v1.NodeSignatureH\x00R\rnodeSignature\x12@\n\x0b\x63hain_event\x18\xd4\x0f \x01(\x0b\x32\x1c.vega.commands.v1.ChainEventH\x00R\nchainEvent\x12\\\n\x15key_rotate_submission\x18\xd5\x0f \x01(\x0b\x32%.vega.commands.v1.KeyRotateSubmissionH\x00R\x13keyRotateSubmission\x12\x62\n\x17state_variable_proposal\x18\xd6\x0f \x01(\x0b\x32'.vega.commands.v1.StateVariableProposalH\x00R\x15stateVariableProposal\x12X\n\x13validator_heartbeat\x18\xd7\x0f \x01(\x0b\x32$.vega.commands.v1.ValidatorHeartbeatH\x00R\x12validatorHeartbeat\x12u\n\x1e\x65thereum_key_rotate_submission\x18\xd8\x0f \x01(\x0b\x32-.vega.commands.v1.EthereumKeyRotateSubmissionH\x00R\x1b\x65thereumKeyRotateSubmission\x12h\n\x19protocol_upgrade_proposal\x18\xd9\x0f \x01(\x0b\x32).vega.commands.v1.ProtocolUpgradeProposalH\x00R\x17protocolUpgradeProposal\x12O\n\x10issue_signatures\x18\xda\x0f \x01(\x0b\x32!.vega.commands.v1.IssueSignaturesH\x00R\x0fissueSignatures\x12_\n\x16oracle_data_submission\x18\xb9\x17 \x01(\x0b\x32&.vega.commands.v1.OracleDataSubmissionH\x00R\x14oracleDataSubmissionB\t\n\x07\x63ommandJ\x06\x08\xd1\x0f\x10\xd2\x0f\x42\x31Z/code.vegaprotocol.io/vega/protos/vega/wallet/v1b\x06proto3"
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "vega.wallet.v1.wallet_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"Z/code.vegaprotocol.io/vega/protos/vega/wallet/v1"
    )
    _SUBMITTRANSACTIONREQUEST._serialized_start = 153
    _SUBMITTRANSACTIONREQUEST._serialized_end = 2512
# @@protoc_insertion_point(module_scope)



from . import data_pb2 as data
from . import data_pb2_grpc as data_grpc
from . import spec_pb2 as spec
from . import spec_pb2_grpc as spec_grpc

__all__ = [
    "data",
    "data_grpc",
    "spec",
    "spec_grpc",
]



from __future__ import annotations

import numpy as np
from collections import namedtuple
from typing import List, Optional, Union
from numpy.typing import ArrayLike
from vega_sim.api.data import Order
from vega_sim.scenario.common.agents import MomentumTrader

from vega_sim.scenario.ideal_market_maker_v2.utils.strategy import (
    A_S_MMmodel,
    GLFT_approx,
)
from vega_sim.environment import VegaState
from vega_sim.environment.agent import StateAgentWithWallet
from vega_sim.null_service import VegaServiceNull
from vega_sim.proto.vega import markets as markets_protos, vega as vega_protos
from vega_sim.service import PeggedOrder

WalletConfig = namedtuple("WalletConfig", ["name", "passphrase"])

# Set up parties in the market/ Submit liquidity provision
MM_WALLET = WalletConfig("mm", "pin")

# Send selling/buying MOs to hit LP orders
TRADER_WALLET = WalletConfig("trader", "trader")

BACKGROUND_MARKET = WalletConfig("market", "market")

MOMENTUM_WALLET = WalletConfig("momentum", "momentum")

# Pass opening auction
AUCTION1_WALLET = WalletConfig("AUCTION1", "AUCTION1pass")
AUCTION2_WALLET = WalletConfig("AUCTION2", "AUCTION2pass")

# Terminate the market and send settlment price
TERMINATE_WALLET = WalletConfig("FJMKnwfZdd48C8NqvYrG", "bY3DxwtsCstMIIZdNpKs")

# informed trader wallet
INFORMED_WALLET = WalletConfig("INFORMED", "INFORMEDpass")


class OptimalMarketMaker(StateAgentWithWallet):
    def __init__(
        self,
        wallet_name: str,
        wallet_pass: str,
        terminate_wallet_name: str,
        terminate_wallet_pass: str,
        price_process: List[float],
        initial_asset_mint: float = 1000000,
        spread: float = 0.002,
        num_steps: int = 180,
        limit_order_size: float = 10,
        market_order_arrival_rate: float = 5,
        kappa: float = 500,
        inventory_upper_boundary: int = 20,
        inventory_lower_boundary: int = -20,
        terminal_penalty_parameter: float = 10**-4,
        running_penalty_parameter: float = 5 * 10**-6,
        asset_decimal: int = 5,
        market_decimal: int = 5,
        market_position_decimal: int = 2,
        market_name: str = None,
        asset_name: str = None,
        set_up_market: bool = True,
        commitment_amount: float = 6000,
        settlement_price: Optional[float] = None,
        tag: str = "",
    ):
        super().__init__(wallet_name + str(tag), wallet_pass)
        self.terminate_wallet_name = terminate_wallet_name + str(tag)
        self.terminate_wallet_pass = terminate_wallet_pass

        self.price_process = price_process
        self.spread = spread
        self.time = num_steps
        self.Lambda = market_order_arrival_rate
        self.kappa = kappa
        self.limit_order_size = limit_order_size
        self.q_upper = inventory_upper_boundary
        self.q_lower = inventory_lower_boundary
        self.alpha = terminal_penalty_parameter
        self.phi = running_penalty_parameter
        self.adp = asset_decimal
        self.mdp = market_decimal
        self.market_position_decimal = market_position_decimal
        self.commitment_amount = commitment_amount
        self.settlement_price = settlement_price
        self.initial_asset_mint = initial_asset_mint

        self.current_step = 0

        self.tag = tag
        self.set_up_market = set_up_market

        self.market_name = f"ETH:USD_{self.tag}" if market_name is None else market_name
        self.asset_name = f"tDAI{self.tag}" if asset_name is None else asset_name

        self.long_horizon_estimate = num_steps >= 200

        if not self.long_horizon_estimate:
            self.optimal_bid, self.optimal_ask, _ = A_S_MMmodel(
                T=self.time / 60 / 24 / 365.25,
                dt=1 / 60 / 24 / 365.25,
                length=self.time + 1,
                q_upper=self.q_upper,
                q_lower=self.q_lower,
                mdp=self.mdp,
                kappa=self.kappa,
                Lambda=self.Lambda,
                alpha=self.alpha,
                phi=self.phi,
            )
        else:
            self.optimal_bid, self.optimal_ask = GLFT_approx(
                q_upper=self.q_upper,
                q_lower=self.q_lower,
                kappa=self.kappa,
                Lambda=self.Lambda,
                alpha=self.alpha,
                phi=self.phi,
            )

    def finalise(self):
        if self.settlement_price:
            self.vega.settle_market(
                self.terminate_wallet_name, self.settlement_price, self.market_id
            )
            self.current_step += 1

    def initialise(self, vega: VegaServiceNull):
        # Initialise wallet for LP/ Settle Party
        super().initialise(vega=vega)
        self.vega.create_wallet(self.terminate_wallet_name, self.terminate_wallet_pass)

        # Faucet vega tokens
        self.vega.wait_fn(10)
        self.vega.wait_for_total_catchup()
        self.vega.mint(
            self.wallet_name,
            asset="VOTE",
            amount=1e4,
        )
        self.vega.wait_fn(10)
        self.vega.wait_for_total_catchup()
        if self.set_up_market:
            # Create asset
            self.vega.create_asset(
                self.wallet_name,
                name=self.asset_name,
                symbol=self.asset_name,
                decimals=self.adp,
            )
            self.vega.wait_fn(5)
            self.vega.wait_for_total_catchup()
        # Get asset id
        self.asset_id = self.vega.find_asset_id(symbol=self.asset_name)
        # Top up asset
        self.vega.mint(
            self.wallet_name,
            asset=self.asset_id,
            amount=self.initial_asset_mint,
        )
        self.vega.wait_fn(10)
        self.vega.wait_for_total_catchup()

        if self.set_up_market:
            self.vega.update_network_parameter(
                self.wallet_name,
                "market.liquidity.minimum.probabilityOfTrading.lpOrders",
                "1e-6",
            )

            self.vega.wait_for_total_catchup()
            self.vega.update_network_parameter(
                self.wallet_name,
                "market.liquidity.stakeToCcySiskas",
                "0.001",
            )

            self.vega.wait_for_datanode_sync()

            # Set up a future market
            self.vega.create_simple_market(
                market_name=self.market_name,
                proposal_wallet=self.wallet_name,
                settlement_asset_id=self.asset_id,
                termination_wallet=self.terminate_wallet_name,
                market_decimals=self.mdp,
                position_decimals=self.market_position_decimal,
                future_asset=self.asset_name,
            )
            self.vega.wait_for_total_catchup()

        # Get market id
        self.market_id = self.vega.find_market_id(name=self.market_name)

        vega.submit_liquidity(
            wallet_name=self.wallet_name,
            market_id=self.market_id,
            commitment_amount=self.commitment_amount,
            fee=0.0015,
            buy_specs=[("PEGGED_REFERENCE_BEST_BID", 5, 1)],
            sell_specs=[("PEGGED_REFERENCE_BEST_ASK", 5, 1)],
            is_amendment=False,
        )

    def optimal_strategy(self, current_position):
        if current_position >= self.q_upper:
            current_bid_depth = (
                self.optimal_bid[self.current_step, 0]
                if not self.long_horizon_estimate
                else self.optimal_bid[0]
            )
            current_ask_depth = (
                1 / 10**self.mdp
            )  # Sell for the smallest possible amount above mid
        elif current_position <= self.q_lower:
            current_bid_depth = (
                1 / 10**self.mdp
            )  # Buy for the smallest possible amount below mid
            current_ask_depth = (
                self.optimal_ask[self.current_step, -1]
                if not self.long_horizon_estimate
                else self.optimal_ask[-1]
            )
        else:
            current_bid_depth = (
                self.optimal_bid[
                    self.current_step, int(self.q_upper - 1 - current_position)
                ]
                if not self.long_horizon_estimate
                else self.optimal_bid[int(self.q_upper - 1 - current_position)]
            )
            current_ask_depth = (
                self.optimal_ask[
                    self.current_step, int(self.q_upper - current_position)
                ]
                if not self.long_horizon_estimate
                else self.optimal_ask[int(self.q_upper - current_position)]
            )

        return current_bid_depth, current_ask_depth

    def step(self, vega_state: VegaState):
        self.current_step += 1

        # Each step, MM posts optimal bid/ask depths
        position = self.vega.positions_by_market(
            wallet_name=self.wallet_name, market_id=self.market_id
        )

        current_position = int(position[0].open_volume) if position else 0
        self.bid_depth, self.ask_depth = self.optimal_strategy(current_position)

        buy_order, sell_order = None, None

        for order in (
            vega_state.market_state.get(self.market_id, {})
            .orders.get(self.vega.wallet.public_key(self.wallet_name), {})
            .values()
        ):
            if order.side == vega_protos.SIDE_BUY:
                buy_order = order
            else:
                sell_order = order

        self._place_orders(
            buy_offset=self.bid_depth - self.spread / 2,
            sell_offset=self.ask_depth - self.spread / 2,
            volume=self.limit_order_size,
            buy_order=buy_order,
            sell_order=sell_order,
        )

    def _place_orders(
        self,
        buy_offset: float,
        sell_offset: float,
        volume: float,
        buy_order: Optional[str] = None,
        sell_order: Optional[str] = None,
    ):
        self._place_or_amend_order(
            offset=buy_offset,
            volume=volume,
            order=buy_order,
            side=vega_protos.SIDE_BUY,
        )

        self._place_or_amend_order(
            offset=sell_offset,
            volume=volume,
            order=sell_order,
            side=vega_protos.SIDE_SELL,
        )

    def _place_or_amend_order(
        self,
        offset: float,
        volume: float,
        side: vega_protos.Side,
        order: Optional[Order] = None,
    ):
        is_buy = side in ["SIDE_BUY", vega_protos.SIDE_BUY]
        reference = (
            vega_protos.PEGGED_REFERENCE_BEST_BID
            if is_buy
            else vega_protos.PEGGED_REFERENCE_BEST_ASK
        )

        if order is None:
            self.vega.submit_order(
                trading_wallet=self.wallet_name,
                market_id=self.market_id,
                pegged_order=PeggedOrder(reference=reference, offset=offset),
                side=side,
                volume=volume,
                order_type=vega_protos.Order.Type.TYPE_LIMIT,
                wait=False,
                time_in_force=vega_protos.Order.TimeInForce.TIME_IN_FORCE_GTC,
            )
        else:
            self.vega.amend_order(
                trading_wallet=self.wallet_name,
                market_id=self.market_id,
                order_id=order.id,
                pegged_reference=reference,
                pegged_offset=offset,
                volume_delta=volume - order.size,
            )



from typing import Optional
import numpy as np


def regime_change_random_walk(
    num_steps: int = 100,
    random_state: Optional[np.random.RandomState] = None,
    sigma_pre: float = 1,
    sigma_post: float = 2,
    drift_pre: float = 0,
    drift_post: float = -1,
    starting_price: float = 100,
    break_point: int = 50,
    decimal_precision: Optional[int] = None,
    trim_to_min: Optional[float] = None,
):
    random_state_set = random_state is not None
    random_state = random_state if random_state_set else np.random.RandomState()

    S = np.zeros(num_steps)
    S[0] = starting_price

    for _ in range(100):
        dW = random_state.randn(num_steps)
        # Simulate external midprice
        for i in range(1, num_steps):
            sigma = sigma_pre if i < break_point else sigma_post
            drift = drift_pre if i < break_point else drift_post
            S[i] = S[i - 1] + drift + sigma * dW[i]

        # market decimal place
        if decimal_precision:
            S = np.round(S, decimal_precision)

        # If random state is passed then error if it generates a negative price
        # Otherwise retry with a new seed

        if trim_to_min is not None:
            S[S < trim_to_min] = trim_to_min

        if (S > 0).all():
            break
        else:
            if random_state_set:
                raise Exception(
                    "Negative price generated with current random seed. Please try"
                    " another or don't specify one"
                )
            random_state = np.random.RandomState()
    if (S < 0).any():
        raise Exception("No valid price series generated after 100 attempts")
    return S



from dataclasses import dataclass
import numpy as np
from collections import namedtuple, defaultdict
from typing import List, Tuple
import os
from functools import partial
from tqdm import tqdm

import torch
import torch.nn as nn
from torch.distributions.categorical import Categorical
from vega_sim.reinforcement.agents.learning_agent import AbstractAction, LearningAgent

import pickle

from vega_sim.reinforcement.networks import (
    Softmax,
    FFN,
    FFN_Params_Normal,
    FFN_Q,
)
from vega_sim.reinforcement.helpers import apply_funcs, to_torch, toggle
from vega_sim.reinforcement.la_market_state import LAMarketState, states_to_sarsa

from vega_sim.reinforcement.distributions import (
    lognorm_sample,
    lognorm_logprob,
    reg_policy,
)

from vega_sim.api.helpers import num_from_padded_int

from vega_sim.environment import VegaState
from vega_sim.environment.agent import StateAgentWithWallet
from vega_sim.null_service import VegaServiceNull
from vega_sim.proto.vega import markets as markets_protos

WalletConfig = namedtuple("WalletConfig", ["name", "passphrase"])

# Send selling/buying MOs to hit LP orders
WALLET = WalletConfig("learner", "learner")


@dataclass
class Action(AbstractAction):
    buy: bool
    sell: bool
    volume: float


@dataclass
class SoftActionWithVol:
    z_sell: torch.Tensor  # sample from N(mu_sell, sigma_sell) to generate selling volume from lognormal
    z_buy: torch.Tensor  # sample from N(mu_buy, sigma_buy) to generate buying volume from lognormal
    c: torch.Tensor  # sell / buy / do nothing probabilities if training. If simulating, value in {0,1,2} indicating sampled action
    mu: torch.Tensor  # [mu_sell, mu_buy]
    sigma: torch.Tensor  # [sigma_sell, sigma_buy]
    volume_sell: torch.Tensor  # sampled volume sell
    volume_buy: torch.Tensor  # sampled volume buy

    def unravel(self):
        return (
            self.z_sell,
            self.z_buy,
            self.c,
            self.mu,
            self.sigma,
            self.volume_sell,
            self.volume_buy,
        )


class LearningAgentWithVol(LearningAgent):
    def __init__(
        self,
        device: str,
        logfile_pol_imp: str,
        logfile_pol_eval: str,
        logfile_pnl: str,
        discount_factor: float,
        num_levels: int,
        wallet_name: str,
        wallet_pass: str,
        market_name: str,
        initial_balance: int,
        position_decimals: int,
        inventory_penalty: float = 0.05,
    ):
        super().__init__(
            device=device,
            logfile_pol_imp=logfile_pol_imp,
            logfile_pol_eval=logfile_pol_eval,
            logfile_pnl=logfile_pnl,
            discount_factor=discount_factor,
            num_levels=num_levels,
            wallet_name=wallet_name,
            wallet_pass=wallet_pass,
            market_name=market_name,
            initial_balance=initial_balance,
            position_decimals=position_decimals,
            inventory_penalty=inventory_penalty,
        )

        # Dimensions of state and action
        self.num_levels = num_levels
        self.state_dim = 6 + 4 * self.num_levels  # from MarketState
        action_discrete_dim = 3

        # NN and optimizer for Q func
        self.q_func = FFN_Q(
            state_dim=self.state_dim,
        )
        self.optimizer_q = torch.optim.RMSprop(self.q_func.parameters(), lr=0.001)

        # NN for policy and its optimizer
        self.policy_discr = FFN(
            sizes=[self.state_dim, 1024, 1024, 1024, action_discrete_dim],
            activation=nn.Tanh,
            output_activation=Softmax,
        )  # this network decides whether to buy/sell/do nothing

        # NN for volume
        self.policy_volume = FFN_Params_Normal(
            n_in=self.state_dim,
            n_distr=2,
            hidden_sizes=[32],
        )

        # And the optimizer needs to include this too
        self.optimizer_pol = torch.optim.RMSprop(
            list(self.policy_volume.parameters())
            + list(self.policy_discr.parameters()),
            lr=0.001,
        )

    def move_to_device(self):
        self.q_func.to(self.device)
        self.policy_volume.to(self.device)
        self.policy_discr.to(self.device)

    def move_to_cpu(self):
        self.q_func.to("cpu")
        self.policy_volume.to("cpu")
        self.policy_discr.to("cpu")

    def _update_memory(
        self,
        state: LAMarketState,
        action: Action,
        reward: float,
        next_state: LAMarketState,
    ):
        """
        Ensure what is the input?
        """
        self.memory["state"].append(state.to_array())
        # action_discrete = 0 if sell
        # action_discrete = 1 if buy
        # action_discrete = 2 if they do not do anything
        if action.sell:
            self.memory["action_discrete"].append([0])
        elif action.buy:
            self.memory["action_discrete"].append([1])
        else:
            self.memory["action_discrete"].append([2])

        self.memory["action_volume"].append([action.volume])
        self.memory["reward"].append([reward])
        if next_state is not np.nan:
            self.memory["next_state"].append(next_state.to_array())
        else:
            self.memory["next_state"].append(np.nan * np.ones_like(state.to_array()))

        return 0

    def create_dataloader(self, batch_size):
        """
        creates dataset and dataloader for training.
        """
        to_torch_device = partial(to_torch, device=self.device)
        dataset_state = apply_funcs(
            value=self.memory["state"], funcs=(np.stack, to_torch_device)
        )
        dataset_action_discrete = apply_funcs(
            value=self.memory["action_discrete"],
            funcs=(np.array, partial(to_torch, dtype=torch.int64, device=self.device)),
        )
        dataset_action_volume = apply_funcs(
            value=self.memory["action_volume"], funcs=(np.array, to_torch_device)
        )
        dataset_reward = apply_funcs(
            value=self.memory["reward"], funcs=(np.array, to_torch_device)
        )
        dataset_next_state = apply_funcs(
            value=self.memory["next_state"], funcs=(np.stack, to_torch_device)
        )

        dataset = torch.utils.data.TensorDataset(
            *(
                dataset_state,
                dataset_action_discrete,
                dataset_action_volume,
                dataset_reward,
                dataset_next_state,
            )
        )
        dataloader = torch.utils.data.DataLoader(
            dataset, batch_size=batch_size, shuffle=True
        )
        return dataloader

    def empty_action(self) -> AbstractAction:
        return Action(True, True, 0.0)

    def step(self, vega_state: VegaState):
        learning_state = self.state(self.vega)
        self.step_num += 1
        self.latest_action = self._step(learning_state)
        self.latest_state = learning_state

        if learning_state.full_balance <= 0:
            return
        if learning_state.market_in_auction:
            return

        if self.latest_action.buy or self.latest_action.sell:
            try:
                self.vega.submit_market_order(
                    trading_wallet=self.wallet_name,
                    market_id=self.market_id,
                    side="SIDE_BUY" if self.latest_action.buy else "SIDE_SELL",
                    volume=self.latest_action.volume,
                    wait=False,
                    fill_or_kill=False,
                )
            except Exception as e:
                print(e)

    def _step(self, vega_state: LAMarketState) -> Action:
        # learned policy
        state = vega_state.to_array().reshape(1, -1)  # adding batch_dimension
        state = torch.from_numpy(state).float()  # .to(self.device)

        with torch.no_grad():
            soft_action = self.sample_action(state=state, sim=True)
        choice = int(soft_action.c.item())
        if choice == 0:  # choice = 0 --> sell
            volume = soft_action.volume_sell.item() * 10 ** (-self.position_decimals)
        elif choice == 1:  # choice = 1 --> buy
            volume = soft_action.volume_buy.item() * 10 ** (-self.position_decimals)
        else:
            volume = 0  # choice=2, hence do nothing, hence volume is irrelevant
        return Action(buy=choice == 0, sell=choice == 1, volume=volume)

    def _step_heuristic(self, vega_state: LAMarketState) -> Action:
        volume = 0.0
        if (
            vega_state.position <= 0
            and vega_state.ask_prices[0] < vega_state.next_price
        ):
            choice = 0  # buy
            volume = 0.01
        elif (
            vega_state.position >= 0
            and vega_state.bid_prices[0] > vega_state.next_price
        ):
            choice = 1  # sell
            volume = 0.01
        else:
            choice = 2  # do nothing

        return Action(buy=choice == 0, sell=choice == 1, volume=volume)

    def sample_action(
        self, state: torch.Tensor, sim: bool = True, evaluate: bool = False
    ):
        """
        Sample an action.

        Returns
        -------
        z_sell: torch.Tensor
            samples from N(0,1) used to sample volume_sell from a lognormal
        z_buy: torch.Tensor
            samples from N(0,1) used to sample volume_buy from a lognormal
        mu: torch.Tensor
            Tensor of shape (batch_size,2). Each column is mu_{sell} and mu_{buy} of lognormal
        sigma: torch.Tensor
            Tensor of shape (batch_size, 2). Each column is sigma_{sell} and sigma_{buy} of lognormal
        c: torch.Tensor
            if sim = False, c is a tensor of shape (batch_size, 3) returning the probs of {sell, buy, do nothign}
            if sim = True, c is a tensor of shape (batch_size, 1) returning the sampled action from {sell, buy, do nothing} (i.e. c is filled with values from {0,1,2})
        mu: torch.Tensor
            Tensor of shape
        sigma: torch.Tensor
            Tensor of shape
        volume_sell: torch.Tensor
            Tensor of shape
        volume_buy: torch.Tensor
            Tensor of shape
        """
        probs = self.policy_discr(
            state
        )  # this is the FFN returning probabilities for the 3 actions
        mu, sigma = self.policy_volume(
            state
        )  # this is the FFN returning statistics for volume distribution
        z_sell, volume_sell = lognorm_sample(mu=mu[:, 0], sigma=sigma[:, 0])
        z_buy, volume_buy = lognorm_sample(mu=mu[:, 1], sigma=sigma[:, 1])
        if sim:
            # We are simulating, hence we sample from discr action
            m = Categorical(probs)
            c = m.sample()
        else:
            c = probs
        if evaluate:
            c = torch.max(probs, 1, keepdim=True)[1]

        return SoftActionWithVol(
            z_sell=z_sell,
            z_buy=z_buy,
            c=c,
            mu=mu,
            sigma=sigma,
            volume_sell=volume_sell,
            volume_buy=volume_buy,
        )

    def policy_eval(
        self,
        batch_size: int,
        n_epochs: int,
    ):
        toggle(self.policy_discr, to=False)
        toggle(self.policy_volume, to=False)
        toggle(self.q_func, to=True)

        dataloader = self.create_dataloader(batch_size=batch_size)

        pbar = tqdm(total=n_epochs)
        for epoch in range(n_epochs):
            for (
                i,
                (
                    batch_state,
                    batch_action_discrete,
                    batch_action_volume,
                    batch_reward,
                    batch_next_state,
                ),
            ) in enumerate(dataloader):
                next_state_terminal = torch.isnan(
                    batch_next_state
                ).float()  # shape (batch_size, dim_state)
                batch_next_state[next_state_terminal.eq(True)] = batch_state[
                    next_state_terminal.eq(True)
                ]
                self.optimizer_q.zero_grad()
                # differentiate between sell and buy volumes for the q_func
                volume_sell = batch_action_volume.clone()
                volume_sell[batch_action_discrete.ne(0)] = 0
                volume_buy = batch_action_volume.clone()
                volume_buy[batch_action_discrete.ne(1)] = 0

                pred = torch.gather(
                    self.q_func(batch_state, volume_sell, volume_buy),
                    dim=1,
                    index=batch_action_discrete,
                )

                with torch.no_grad():
                    v = self.v_func(batch_next_state)
                    target = (
                        batch_reward
                        + (1 - next_state_terminal.mean(1, keepdim=True))
                        * self.discount_factor
                        * v
                    )
                loss = torch.pow(pred - target, 2).mean()
                loss.backward()
                self.optimizer_q.step()
            self.losses["q"].append(loss.item())
            # logging loss
            with open(self.logfile_pol_eval, "a") as f:
                f.write(
                    "{},{:.2e},{:.3f},{:.3f}\n".format(
                        epoch + self.lerningIteration * n_epochs,
                        loss.item(),
                        self.coefH_discr,
                        self.coefH_cont,
                    )
                )
            pbar.update(1)
        return 0

    def v_func(self, state, n_mc=1000):
        """
        v(x) = E[q(x,A)] = E[q(x,A)|C=0]p(C=0) + E[q(x,A)|C=1]p(C=1) + E[q(x)|C=2]p(C=2)

        if C==0 --> sell,
        if C==1 --> buy,
        if C==2 --> does nothing,
        Parameters
        ----------
        state: torch.Tensor
            State. Tensor of shape (batch_size,
        n_mc: torch.Tensor
            Number of Monte Carlo samples to calculate
        """
        batch_size = state.shape[0]
        # MONTE CARLO to approximate expectations
        state_mc = state.repeat(n_mc, 1)
        soft_action = self.sample_action(state_mc, sim=False)
        z_sell, z_buy, c, mu, sigma, volume_sell, volume_buy = soft_action.unravel()

        q = self.q_func(state_mc, volume_sell, volume_buy)
        v = c[:, 0] * (
            q[:, 0]
            - self.coefH_discr * torch.log(c[:, 0])
            - self.coefH_cont
            * lognorm_logprob(z=z_sell, mu=mu[:, 0], sigma=sigma[:, 0])
        )
        v += c[:, 1] * (
            q[:, 1]
            - self.coefH_discr * torch.log(c[:, 1])
            - self.coefH_cont * lognorm_logprob(z=z_buy, mu=mu[:, 1], sigma=sigma[:, 1])
        )
        v += c[:, 2] * (q[:, 2] - self.coefH_discr * torch.log(c[:, 2]))
        v = v.reshape(n_mc, batch_size, -1).mean(0)  # average of Monte Carlo samples
        self.q_func.train()
        return v

    def D_KL(self, state, n_mc):
        """
        KL divergence between pi(.|x) and the unnormalised density exp(Q(x,.)),
        where pi(.|x) is a LogNormal distribution
        """
        batch_size = state.shape[0]
        # MONTE CARLO to approximate expectations
        state_mc = state.repeat(n_mc, 1)
        soft_action = self.sample_action(state_mc, sim=False)
        z_sell, z_buy, c, mu, sigma, volume_sell, volume_buy = soft_action.unravel()

        q = self.q_func(state_mc, volume_sell, volume_buy)
        d_kl = c[:, 0] * (
            self.coefH_discr * torch.log(c[:, 0])
            + self.coefH_cont
            * lognorm_logprob(z=z_sell, mu=mu[:, 0], sigma=sigma[:, 0])
            - q[:, 0]
        )
        d_kl += c[:, 1] * (
            self.coefH_discr * torch.log(c[:, 1])
            + self.coefH_cont * lognorm_logprob(z=z_buy, mu=mu[:, 1], sigma=sigma[:, 1])
            - q[:, 1]
        )
        d_kl += c[:, 2] * (self.coefH_discr * torch.log(c[:, 2]) - q[:, 2])

        # regularisation. Since the action space is not compact, then Q(x,a) could potentially explode.
        # To avoid this, I force the parameters of the lognormal not to be very far away from (0,1)
        reg = c[:, 0] * reg_policy(z_sell, mu=mu[:, 0], sigma=sigma[:, 0])
        reg += c[:, 1] * reg_policy(z_buy, mu=mu[:, 1], sigma=sigma[:, 1])

        d_kl = (
            d_kl.reshape(n_mc, batch_size, -1).mean(0).mean()
            + 0.5 * self.coefH_cont * reg.reshape(n_mc, batch_size, -1).mean(0).mean()
        )  # Average of Monte Carlo samples. Doing one extra unnecessary step for clarity

        return d_kl

    def policy_improvement(self, batch_size: int, n_epochs: int):
        toggle(self.policy_discr, to=True)
        toggle(self.policy_volume, to=True)
        toggle(self.q_func, to=False)

        dataloader = self.create_dataloader(batch_size=batch_size)

        pbar = tqdm(total=n_epochs)
        for epoch in range(n_epochs):
            for i, (batch_state, _, _, _, _) in enumerate(dataloader):
                self.optimizer_pol.zero_grad()
                d_kl = self.D_KL(batch_state, n_mc=100).mean()
                d_kl.backward()
                # nn.utils.clip_grad_norm_(self.policy_volume.parameters(), max_norm=1.)
                self.optimizer_pol.step()
            self.losses["d_kl"].append(d_kl.item())
            with open(self.logfile_pol_imp, "a") as f:
                f.write(
                    "{},{:.4f}\n".format(
                        epoch + n_epochs * self.lerningIteration, d_kl.item()
                    )
                )
            pbar.update(1)

        # update the coefficients for the next run
        self.coefH_discr = max(self.coefH_discr * 0.99, 0.05)
        self.coefH_cont = max(self.coefH_cont * 0.99, 0.05)
        self.lerningIteration += 1
        return 0

    def save(self, results_dir: str):
        filename = os.path.join(results_dir, "agent.pth.tar")
        d = {
            "losses": self.losses,
            "q": self.q_func.state_dict(),
            "policy_discr": self.policy_discr.state_dict(),
            "policy_volume": self.policy_volume.state_dict(),
            "iteration": self.lerningIteration,
            "coefH_discr": self.coefH_discr,
            "coefH_cont": self.coefH_cont,
        }
        torch.save(d, filename)

        filename_for_memory = os.path.join(results_dir, "memory.pickle")
        with open(filename_for_memory, "wb") as f:
            pickle.dump(self.memory, f)

    def load(self, results_dir: str):
        filename = os.path.join(results_dir, "agent.pth.tar")
        d = torch.load(filename, map_location="cpu")
        self.q_func.load_state_dict(d["q"])
        self.policy_discr.load_state_dict(d["policy_discr"])
        self.policy_volume.load_state_dict(d["policy_volume"])
        self.lerningIteration = d["iteration"]
        self.coefH_discr = d["coefH_discr"]
        self.coefH_cont = d["coefH_cont"]

        filename_for_memory = os.path.join(results_dir, "memory.pickle")
        with open(filename_for_memory, "rb") as f:
            memory = pickle.load(f)
        self.memory = memory



import argparse
import logging
import numpy as np
from typing import Any, Callable, List, Optional
from vega_sim.environment.agent import Agent

from vega_sim.scenario.scenario import Scenario
from vega_sim.scenario.common.utils.price_process import random_walk
from vega_sim.environment.environment import MarketEnvironmentWithState
from vega_sim.null_service import VegaServiceNull
from vega_sim.scenario.constants import Network
from vega_sim.scenario.ideal_market_maker_v2.agents import (
    MM_WALLET,
    TERMINATE_WALLET,
    TRADER_WALLET,
    BACKGROUND_MARKET,
    AUCTION1_WALLET,
    AUCTION2_WALLET,
    INFORMED_WALLET,
    OptimalMarketMaker,
)
from vega_sim.scenario.common.agents import (
    MarketOrderTrader,
    BackgroundMarket,
    OpenAuctionPass,
    InformedTrader,
)


class IdealMarketMaker(Scenario):
    def __init__(
        self,
        num_steps: int = 120,
        random_agent_ordering: bool = True,
        market_decimal: int = 5,
        asset_decimal: int = 5,
        market_position_decimal: int = 2,
        market_name: str = None,
        asset_name: str = None,
        initial_asset_mint: float = 1000000,
        initial_price: Optional[float] = None,
        sigma: float = 1,
        kappa: float = 1,
        q_upper: int = 20,
        q_lower: int = -20,
        alpha: float = 10**-4,
        phi: float = 5 * 10**-6,
        market_marker_limit_order_size: float = 10,
        lp_commitamount: float = 200000,
        spread: int = 2,
        block_size: int = 1,
        block_length_seconds: int = 1,
        state_extraction_freq: int = 1,
        buy_intensity: float = 5,
        sell_intensity: float = 5,
        backgroundmarket_tick_spacing: float = 0.002,
        backgroundmarket_number_levels_per_side: int = 20,
        step_length_seconds: int = 1,
        opening_auction_trade_amount: float = 1,
        state_extraction_fn: Optional[
            Callable[[VegaServiceNull, List[Agent]], Any]
        ] = None,
        pause_every_n_steps: Optional[int] = None,
        price_process_fn: Optional[Callable[[None], List[float]]] = None,
        market_order_trader_base_order_size: float = 1,
        settle_at_end: bool = True,
        proportion_taken: float = 0.8,
    ):
        self.num_steps = num_steps
        self.random_agent_ordering = random_agent_ordering
        self.market_decimal = market_decimal
        self.asset_decimal = asset_decimal
        self.market_position_decimal = market_position_decimal
        self.initial_price = initial_price
        self.sigma = sigma
        self.kappa = kappa
        self.q_upper = q_upper
        self.q_lower = q_lower
        self.alpha = alpha
        self.phi = phi
        self.spread = spread / 10**self.market_decimal
        self.block_size = block_size
        self.block_length_seconds = block_length_seconds
        self.state_extraction_freq = state_extraction_freq
        self.step_length_seconds = step_length_seconds
        self.state_extraction_fn = state_extraction_fn
        self.buy_intensity = buy_intensity
        self.sell_intensity = sell_intensity
        self.pause_every_n_steps = pause_every_n_steps
        self.lp_commitamount = lp_commitamount
        self.initial_asset_mint = initial_asset_mint
        self.backgroundmarket_tick_spacing = backgroundmarket_tick_spacing
        self.market_marker_limit_order_size = market_marker_limit_order_size
        self.backgroundmarket_number_levels_per_side = (
            backgroundmarket_number_levels_per_side
        )
        self.market_name = "ETH:USD" if market_name is None else market_name
        self.asset_name = "tDAI" if asset_name is None else asset_name
        self.price_process_fn = price_process_fn
        self.opening_auction_trade_amount = opening_auction_trade_amount
        self.market_order_trader_base_order_size = market_order_trader_base_order_size
        self.settle_at_end = settle_at_end
        self.price_process = None
        self.proportion_taken = proportion_taken

    def _generate_price_process(
        self,
        random_state: Optional[np.random.RandomState] = None,
    ):
        return random_walk(
            num_steps=self.num_steps,
            sigma=self.sigma,
            starting_price=self.initial_price,
            random_state=random_state,
        )

    def set_up_background_market(
        self,
        vega: VegaServiceNull,
        tag: str = "",
        random_state: Optional[np.random.RandomState] = None,
    ) -> MarketEnvironmentWithState:
        # Set up market name and settlement asset
        market_name = self.market_name + f"_{tag}"
        asset_name = self.asset_name + f"_{tag}"

        price_process = (
            self.price_process_fn()
            if self.price_process_fn is not None
            else self._generate_price_process(random_state=random_state)
        )
        self.price_process = price_process

        market_maker = OptimalMarketMaker(
            wallet_name=MM_WALLET.name,
            wallet_pass=MM_WALLET.passphrase,
            terminate_wallet_name=TERMINATE_WALLET.name,
            terminate_wallet_pass=TERMINATE_WALLET.passphrase,
            initial_asset_mint=self.initial_asset_mint,
            price_process=price_process,
            spread=self.spread,
            num_steps=self.num_steps,
            market_order_arrival_rate=self.buy_intensity,
            kappa=self.kappa,
            limit_order_size=self.market_marker_limit_order_size,
            inventory_upper_boundary=self.q_upper,
            inventory_lower_boundary=self.q_lower,
            terminal_penalty_parameter=self.alpha,
            running_penalty_parameter=self.phi,
            asset_decimal=self.asset_decimal,
            market_decimal=self.market_decimal,
            market_position_decimal=self.market_position_decimal,
            market_name=market_name,
            asset_name=asset_name,
            commitment_amount=self.lp_commitamount,
            tag=str(tag),
            settlement_price=price_process[-1] if self.settle_at_end else None,
        )

        trader = MarketOrderTrader(
            wallet_name=TRADER_WALLET.name,
            wallet_pass=TRADER_WALLET.passphrase,
            initial_asset_mint=self.initial_asset_mint,
            market_name=market_name,
            asset_name=asset_name,
            tag=str(tag),
            buy_intensity=self.buy_intensity,
            sell_intensity=self.sell_intensity,
            random_state=random_state,
            base_order_size=self.market_order_trader_base_order_size,
        )

        background_market = BackgroundMarket(
            wallet_name=BACKGROUND_MARKET.name,
            wallet_pass=BACKGROUND_MARKET.passphrase,
            market_name=market_name,
            asset_name=asset_name,
            initial_asset_mint=self.initial_asset_mint,
            price_process=price_process,
            spread=self.spread,
            tick_spacing=self.backgroundmarket_tick_spacing,
            order_distribution_kappa=self.kappa,
            num_levels_per_side=self.backgroundmarket_number_levels_per_side,
            tag=str(tag),
            position_decimals=self.market_position_decimal,
        )

        auctionpass1 = OpenAuctionPass(
            wallet_name=AUCTION1_WALLET.name,
            wallet_pass=AUCTION1_WALLET.passphrase,
            side="SIDE_BUY",
            initial_asset_mint=self.initial_asset_mint,
            initial_price=self.initial_price
            if self.initial_price is not None
            else price_process[0],
            market_name=market_name,
            asset_name=asset_name,
            opening_auction_trade_amount=self.opening_auction_trade_amount,
            tag=str(tag),
        )

        auctionpass2 = OpenAuctionPass(
            wallet_name=AUCTION2_WALLET.name,
            wallet_pass=AUCTION2_WALLET.passphrase,
            side="SIDE_SELL",
            initial_asset_mint=self.initial_asset_mint,
            initial_price=self.initial_price
            if self.initial_price is not None
            else price_process[0],
            market_name=market_name,
            asset_name=asset_name,
            opening_auction_trade_amount=self.opening_auction_trade_amount,
            tag=str(tag),
        )

        info_trader = InformedTrader(
            wallet_name=INFORMED_WALLET.name,
            wallet_pass=INFORMED_WALLET.passphrase,
            price_process=price_process,
            market_name=market_name,
            asset_name=asset_name,
            initial_asset_mint=self.initial_asset_mint,
            proportion_taken=self.proportion_taken,
            tag=str(tag),
        )

        env = MarketEnvironmentWithState(
            agents=[
                market_maker,
                background_market,
                auctionpass1,
                auctionpass2,
                trader,
            ],
            n_steps=self.num_steps,
            random_agent_ordering=self.random_agent_ordering,
            transactions_per_block=self.block_size,
            vega_service=vega,
            state_extraction_freq=self.state_extraction_freq,
            step_length_seconds=self.step_length_seconds,
            block_length_seconds=self.block_length_seconds,
            state_extraction_fn=self.state_extraction_fn,
            pause_every_n_steps=self.pause_every_n_steps,
        )
        return env

    def run_iteration(
        self,
        vega: VegaServiceNull,
        network: Optional[Network] = None,
        pause_at_completion: bool = False,
        run_with_console: bool = False,
        random_state: Optional[np.random.RandomState] = None,
    ):
        env = self.set_up_background_market(
            vega=vega, tag=str(0), random_state=random_state
        )
        result = env.run(
            pause_at_completion=pause_at_completion,
            run_with_console=run_with_console,
        )
        return result


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO if not args.debug else logging.DEBUG)

    step_length = 60 * 60

    scenario = IdealMarketMaker(num_steps=200)

    with VegaServiceNull(
        warn_on_raw_data_access=False,
        run_with_console=True,
        seconds_per_block=40,  # Heuristic
    ) as vega:
        scenario.run_iteration(
            vega=vega,
            pause_at_completion=True,
        )



import requests
from logging import getLogger

BASE_MINT_URL = "{faucet_url}/api/v1/mint"

logger = getLogger(__name__)


def mint(pub_key: str, asset: str, amount: int, faucet_url: str) -> None:
    url = BASE_MINT_URL.format(faucet_url=faucet_url)
    payload = {
        "party": pub_key,
        "amount": str(int(amount)),
        "asset": asset,
    }
    req = requests.post(url, json=payload)
    try:
        req.raise_for_status()
    except Exception as e:
        logger.exception(f"Exception in minting: {req.json()}")
        raise e



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vega/commands/v1/data.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1bvega/commands/v1/data.proto\x12\x10vega.commands.v1"\xe3\x01\n\x14OracleDataSubmission\x12K\n\x06source\x18\x01 \x01(\x0e\x32\x33.vega.commands.v1.OracleDataSubmission.OracleSourceR\x06source\x12\x18\n\x07payload\x18\x02 \x01(\x0cR\x07payload"d\n\x0cOracleSource\x12\x1d\n\x19ORACLE_SOURCE_UNSPECIFIED\x10\x00\x12\x1d\n\x19ORACLE_SOURCE_OPEN_ORACLE\x10\x01\x12\x16\n\x12ORACLE_SOURCE_JSON\x10\x02\x42\x33Z1code.vegaprotocol.io/vega/protos/vega/commands/v1b\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "vega.commands.v1.data_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"Z1code.vegaprotocol.io/vega/protos/vega/commands/v1"
    )
    _ORACLEDATASUBMISSION._serialized_start = 50
    _ORACLEDATASUBMISSION._serialized_end = 277
    _ORACLEDATASUBMISSION_ORACLESOURCE._serialized_start = 177
    _ORACLEDATASUBMISSION_ORACLESOURCE._serialized_end = 277
# @@protoc_insertion_point(module_scope)



"""agents.py

File contains the party config tuples for each agent in the MultiMarket scenario. The
scenario runs a network with three markets each with a configurable number of market
order trader agents.

"""
from collections import namedtuple

PartyConfig = namedtuple("AgentConfig", ["wallet_name", "wallet_pass", "key_name"])

# Set-up wallets and keys for MarketManager agents
MARKET_MANAGERS = {
    "MARKET_A_CREATOR": PartyConfig("market_managers", "pass", "Market A Creator"),
    "MARKET_A_SETTLER": PartyConfig("market_managers", "pass", "Market A Settler"),
    "MARKET_B_CREATOR": PartyConfig("market_managers", "pass", "Market B Creator"),
    "MARKET_B_SETTLER": PartyConfig("market_managers", "pass", "Market B Settler"),
    "MARKET_C_CREATOR": PartyConfig("market_managers", "pass", "Market C Creator"),
    "MARKET_C_SETTLER": PartyConfig("market_managers", "pass", "Market C Settler"),
}

# Set-up wallets and keys for CurvedMarketMaker agents
MARKET_MAKERS = {
    "MARKET_A_MAKER": PartyConfig("market_makers", "pass", "Market A Maker"),
    "MARKET_B_MAKER": PartyConfig("market_makers", "pass", "Market B Maker"),
    "MARKET_C_MAKER": PartyConfig("market_makers", "pass", "Market C Maker"),
}

# Set-up wallets and keys for AuctionPass agents
MARKET_PASSERS = {
    "MARKET_A_PASSER_BID": PartyConfig("market_passers", "pass", "Market A Passer Bid"),
    "MARKET_A_PASSER_ASK": PartyConfig("market_passers", "pass", "Market A Passer Ask"),
    "MARKET_B_PASSER_BID": PartyConfig("market_passers", "pass", "Market B Passer Bid"),
    "MARKET_B_PASSER_ASK": PartyConfig("market_passers", "pass", "Market B Passer Ask"),
    "MARKET_C_PASSER_BID": PartyConfig("market_passers", "pass", "Market C Passer Bid"),
    "MARKET_C_PASSER_ASK": PartyConfig("market_passers", "pass", "Market C Passer Ask"),
}

# Set-up wallets and keys for MarketOrderTrader agents
MARKET_TRADERS = {
    f"MARKET_{i}_TRADER_{str(k).zfill(4)}": PartyConfig(
        "market_traders", "pass", f"Market {i} Trader {str(k).zfill(4)}"
    )
    for i in ["A", "B", "C"]
    for k in range(1000)
}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: blockexplorer/blockexplorer.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from vega.commands.v1 import (
    transaction_pb2 as vega_dot_commands_dot_v1_dot_transaction__pb2,
)
from protoc_gen_openapiv2.options import (
    annotations_pb2 as protoc__gen__openapiv2_dot_options_dot_annotations__pb2,
)


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n!blockexplorer/blockexplorer.proto\x12\x14\x62lockexplorer.api.v1\x1a"vega/commands/v1/transaction.proto\x1a.protoc-gen-openapiv2/options/annotations.proto"+\n\x15GetTransactionRequest\x12\x12\n\x04hash\x18\x01 \x01(\tR\x04hash"]\n\x16GetTransactionResponse\x12\x43\n\x0btransaction\x18\x01 \x01(\x0b\x32!.blockexplorer.api.v1.TransactionR\x0btransaction"\x8e\x02\n\x17ListTransactionsRequest\x12\x14\n\x05limit\x18\x01 \x01(\rR\x05limit\x12\x1b\n\x06\x62\x65\x66ore\x18\x02 \x01(\tH\x00R\x06\x62\x65\x66ore\x88\x01\x01\x12\x19\n\x05\x61\x66ter\x18\x03 \x01(\tH\x01R\x05\x61\x66ter\x88\x01\x01\x12T\n\x07\x66ilters\x18\x04 \x03(\x0b\x32:.blockexplorer.api.v1.ListTransactionsRequest.FiltersEntryR\x07\x66ilters\x1a:\n\x0c\x46iltersEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\t\n\x07_beforeB\x08\n\x06_after"a\n\x18ListTransactionsResponse\x12\x45\n\x0ctransactions\x18\x03 \x03(\x0b\x32!.blockexplorer.api.v1.TransactionR\x0ctransactions"\xe2\x01\n\x0bTransaction\x12\x14\n\x05\x62lock\x18\x01 \x01(\x04R\x05\x62lock\x12\x14\n\x05index\x18\x02 \x01(\rR\x05index\x12\x12\n\x04hash\x18\x03 \x01(\tR\x04hash\x12\x1c\n\tsubmitter\x18\x04 \x01(\tR\tsubmitter\x12\x12\n\x04type\x18\x05 \x01(\tR\x04type\x12\x12\n\x04\x63ode\x18\x06 \x01(\rR\x04\x63ode\x12\x16\n\x06\x63ursor\x18\x07 \x01(\tR\x06\x63ursor\x12\x35\n\x07\x63ommand\x18\x08 \x01(\x0b\x32\x1b.vega.commands.v1.InputDataR\x07\x63ommand2\xfa\x01\n\x14\x42lockExplorerService\x12m\n\x0eGetTransaction\x12+.blockexplorer.api.v1.GetTransactionRequest\x1a,.blockexplorer.api.v1.GetTransactionResponse"\x00\x12s\n\x10ListTransactions\x12-.blockexplorer.api.v1.ListTransactionsRequest\x1a..blockexplorer.api.v1.ListTransactionsResponse"\x00\x42xZ5code.vegaprotocol.io/vega/protos/blockexplorer/api/v1\x92\x41>\x12#\n\x18Vega block explorer APIs2\x07v0.59.0\x1a\x13lb.testnet.vega.xyz*\x02\x01\x02\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "blockexplorer.blockexplorer_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"Z5code.vegaprotocol.io/vega/protos/blockexplorer/api/v1\222A>\022#\n\030Vega block explorer APIs2\007v0.59.0\032\023lb.testnet.vega.xyz*\002\001\002"
    _LISTTRANSACTIONSREQUEST_FILTERSENTRY._options = None
    _LISTTRANSACTIONSREQUEST_FILTERSENTRY._serialized_options = b"8\001"
    _GETTRANSACTIONREQUEST._serialized_start = 143
    _GETTRANSACTIONREQUEST._serialized_end = 186
    _GETTRANSACTIONRESPONSE._serialized_start = 188
    _GETTRANSACTIONRESPONSE._serialized_end = 281
    _LISTTRANSACTIONSREQUEST._serialized_start = 284
    _LISTTRANSACTIONSREQUEST._serialized_end = 554
    _LISTTRANSACTIONSREQUEST_FILTERSENTRY._serialized_start = 475
    _LISTTRANSACTIONSREQUEST_FILTERSENTRY._serialized_end = 533
    _LISTTRANSACTIONSRESPONSE._serialized_start = 556
    _LISTTRANSACTIONSRESPONSE._serialized_end = 653
    _TRANSACTION._serialized_start = 656
    _TRANSACTION._serialized_end = 882
    _BLOCKEXPLORERSERVICE._serialized_start = 885
    _BLOCKEXPLORERSERVICE._serialized_end = 1135
# @@protoc_insertion_point(module_scope)



from abc import ABC
from typing import Any

VEGA_DEFAULT_KEY_NAME = "foo"


class Wallet(ABC):
    def create_wallet(
        self,
        name: str,
        passphrase: str,
    ) -> str:
        """Generates a new wallet from a name - passphrase pair in the given vega service.

        Args:
            name:
                str, The name to use for the wallet
            passphrase:
                str, The passphrase to use when logging in to created wallet in future
        Returns:
            str, login token to use in authenticated requests
        """
        pass

    def login(self, name: str, passphrase: str) -> str:
        """Logs in to existing wallet in the given vega service.

        Args:
            name:
                str, The name of the wallet
            passphrase:
                str, The login passphrase used when creating the wallet
        Returns:
            str, login token to use in authenticated requests
        """
        pass

    def submit_transaction(self, transaction: Any, name: str, transaction_type: str):
        """Submits a transaction to Vega core via wallet

        Args:
            transaction:
                Any, The transaction object which will be submitted
            name:
                str, The name to use for the wallet executing the transaction
            transaction_type:
                str, The name, in underscore case, of the transaction
        """
        pass

    def public_key(self, name: str) -> str:
        """Return the public key associated with a given wallet name.

        Args:
            name:
                str, The name to use for the wallet

        Returns:
            str, public key
        """
        pass



import functools
import vega_sim.proto.vega as vega_protos
from typing import Any, Callable, Dict, List, Optional
from vega_sim.null_service import VegaServiceNull
from vega_sim.environment.agent import Agent
from vega_sim.scenario.common.agents import ExponentialShapedMarketMaker
from vega_sim.scenario.ideal_market_maker.agents import OptimalMarketMaker
from vega_sim.scenario.ideal_market_maker_v2.agents import (
    OptimalMarketMaker as OptimalMarketMakerV2,
)
from vega_sim.scenario.common.agents import (
    MomentumTrader,
    MarketOrderTrader,
    InformedTrader,
)


BASE_IDEAL_MM_CSV_HEADERS = [
    "Time Step",
    "LP: General Account",
    "LP: Margin Account",
    "LP: Margin Rate",
    "LP: Max Margin",
    "LP: Max Locked Capital",
    "LP: Bond Account",
    "LP: GeneralPnl",
    "LP: Return Over LockedCapital",
    "LP: Annualised Return Over LockedCapital",
    "LP: Relative Return",
    "LP: Annualised Relative Return",
    "LP: RealisedPnl",
    "LP: UnrealisedPnl",
    "LP: Position",
    "LP: Bid",
    "LP: Ask",
    "External Midprice",
    "Midprice",
    "Markprice",
    "LP: entry price",
    "InsurancePool",
    "LiquifeeAccount",
    "InfrafeeAccount",
    "Total Traded Notional",
    "Informed Trader Notional",
    "Target Stake",
    "Market Open Interest",
    "Market Trading mode",
    "Market State",
]

LOB_CSV_HEADERS = [
    "Time Step",
    "Order Book Bid Side",
    "Order Book Ask Side",
]


def ideal_market_maker_single_data_extraction(
    additional_data_fns: Optional[
        List[Callable[[VegaServiceNull, List[Agent]], Dict[str, Any]]]
    ] = None,
) -> Callable[[VegaServiceNull, List[Agent]], Dict[str, Any]]:
    return functools.partial(
        _ideal_market_maker_single_data_extraction,
        additional_data_fns=additional_data_fns,
    )


def _ideal_market_maker_single_data_extraction(
    vega: VegaServiceNull,
    agents: List[Agent],
    state_values: List = None,
    additional_data_fns: List[
        Optional[Callable[[VegaServiceNull, List[Agent]], Dict[str, Any]]]
    ] = None,
) -> Dict[str, Any]:
    mm_agent = [
        agent
        for agent in agents
        if isinstance(
            agent,
            (OptimalMarketMakerV2, OptimalMarketMaker, ExponentialShapedMarketMaker),
        )
    ][0]

    # informed_agent = [
    #     agent
    #     for agent in agents
    #     if isinstance(
    #         agent,
    #         (InformedTrader),
    #     )
    # ][0]

    general_lp, margin_lp, bond_lp = vega.party_account(
        wallet_name=mm_agent.wallet_name,
        asset_id=mm_agent.asset_id,
        market_id=mm_agent.market_id,
        key_name=mm_agent.key_name,
    )

    position = vega.positions_by_market(
        wallet_name=mm_agent.wallet_name,
        market_id=mm_agent.market_id,
        key_name=mm_agent.key_name,
    )
    if not position:
        realised_pnl_lp = 0
        unrealised_pnl_lp = 0
        inventory_lp = 0
        entry_price = 0
    else:
        realised_pnl_lp = round(float(position[0].realised_pnl), mm_agent.adp)
        unrealised_pnl_lp = round(float(position[0].unrealised_pnl), mm_agent.adp)
        inventory_lp = float(position[0].open_volume)
        entry_price = float(position[0].average_entry_price) / 10**mm_agent.mdp

    market_state = vega.market_info(market_id=mm_agent.market_id).state
    market_data = vega.market_data(market_id=mm_agent.market_id)
    markprice = float(market_data.mark_price) / 10**mm_agent.mdp
    mid_price = float(market_data.mid_price) / 10**mm_agent.mdp
    trading_mode = market_data.market_trading_mode

    liquifee, insurance = [
        int(account.balance)
        for account in vega.market_accounts(
            asset_id=mm_agent.asset_id, market_id=mm_agent.market_id
        )
    ]

    infrafee = int(
        vega.infrastructure_fee_accounts(asset_id=mm_agent.asset_id)[0].balance
    )
    if hasattr(mm_agent, "adp"):
        infrafee /= 10**mm_agent.adp
    infrafee_rate = float(
        vega.market_info(market_id=mm_agent.market_id).fees.factors.infrastructure_fee
    )
    traded_notional = round(infrafee / infrafee_rate, 3)
    margin_rate = float(margin_lp / (margin_lp + bond_lp + general_lp))

    additional_fns = additional_data_fns if additional_data_fns is not None else []

    if len(state_values) == 0:
        max_locked_capital = margin_lp + bond_lp
        max_margin = margin_lp
    else:
        old_logs = state_values[mm_agent.current_step - 2]
        max_locked_capital = old_logs["LP: Max Locked Capital"]
        max_locked_capital = max(margin_lp + bond_lp, max_locked_capital)
        max_margin = old_logs["LP: Max Margin"]
        max_margin = max(margin_lp, max_margin)

    base_logs = {
        "Time Step": mm_agent.current_step,
        "LP: General Account": general_lp,
        "LP: Margin Account": margin_lp,
        "LP: Margin Rate": margin_rate,
        "LP: Bond Account": bond_lp,
        "LP: GeneralPnl": general_lp
        + margin_lp
        + bond_lp
        - mm_agent.initial_asset_mint,
        "LP: Return Over LockedCapital": (
            general_lp + margin_lp + bond_lp - mm_agent.initial_asset_mint
        )
        / (max_locked_capital + 0.000000001),
        "LP: Annualised Return Over LockedCapital": (
            general_lp + margin_lp + bond_lp - mm_agent.initial_asset_mint
        )
        * 365
        / (max_locked_capital + 0.000000001),
        "LP: Relative Return": (
            general_lp + margin_lp + bond_lp - mm_agent.initial_asset_mint
        )
        / (mm_agent.initial_asset_mint + 0.00000001),
        "LP: Annualised Relative Return": (
            general_lp + margin_lp + bond_lp - mm_agent.initial_asset_mint
        )
        * 365
        / (mm_agent.initial_asset_mint + 0.00000001),
        "LP: Max Margin": max_margin,
        "LP: Max Locked Capital": max_locked_capital,
        "LP: RealisedPnl": realised_pnl_lp,
        "LP: UnrealisedPnl": unrealised_pnl_lp,
        "LP: Position": inventory_lp,
        "LP: Bid": -round(mm_agent.bid_depth, mm_agent.mdp),
        "LP: Ask": round(mm_agent.ask_depth, mm_agent.mdp),
        "External Midprice": mm_agent.price_process[mm_agent.current_step - 1]
        if hasattr(mm_agent, "price_process")
        else None,
        "Midprice": mid_price,
        "Markprice": markprice,
        "LP: entry price": entry_price,
        "InsurancePool": insurance,
        "LiquifeeAccount": liquifee,
        "InfrafeeAccount": infrafee,
        "Total Traded Notional": traded_notional,
        "Informed Trader Notional": traded_notional,
        "Market Trading mode": trading_mode,
        "Market State": market_state,
    }
    for data_fn in additional_fns:
        base_logs.update(data_fn(vega, agents))

    return base_logs


def v1_ideal_mm_additional_data(
    vega: VegaServiceNull,
    agents: List[Agent],
) -> Dict[str, Any]:
    mm_agent = [agent for agent in agents if isinstance(agent, OptimalMarketMaker)][0]

    return {
        "Buying MOs": mm_agent.num_buyMO,
        "Selling MOs": mm_agent.num_sellMO,
        "LOs at bid": mm_agent.num_bidhit,
        "LOs at ask": mm_agent.num_askhit,
    }


def target_stake_additional_data(
    vega: VegaServiceNull,
    agents: List[Agent],
) -> Dict[str, Any]:
    mm_agent = [
        agent
        for agent in agents
        if isinstance(
            agent,
            (OptimalMarketMakerV2, OptimalMarketMaker, ExponentialShapedMarketMaker),
        )
    ][0]
    market_data = vega.market_data(market_id=mm_agent.market_id)
    scaling = 1 / 10 ** mm_agent.adp if hasattr(mm_agent, "adp") else 1

    return {
        "Target Stake": float(market_data.target_stake) * scaling,
    }


def tau_scaling_additional_data(
    vega: VegaServiceNull,
    agents: List[Agent],
) -> Dict[str, Any]:
    mm_agent = [
        agent
        for agent in agents
        if isinstance(
            agent,
            (OptimalMarketMakerV2, OptimalMarketMaker, ExponentialShapedMarketMaker),
        )
    ][0]
    market_data = vega.market_data(market_id=mm_agent.market_id)
    market_info = vega.market_info(market_id=mm_agent.market_id)

    return {
        "Market Open Interest": int(market_data.open_interest)
        / 10**market_info.position_decimal_places
    }


def limit_order_book(
    vega: VegaServiceNull,
    agents: List[Agent],
) -> Dict[str, Any]:
    mm_agent = [
        agent
        for agent in agents
        if isinstance(
            agent,
            (OptimalMarketMakerV2, OptimalMarketMaker, ExponentialShapedMarketMaker),
        )
    ][0]
    order_book = []
    for _, orders in (
        vega.order_status_from_feed(live_only=True).get(mm_agent.market_id, {}).items()
    ):
        order_book += list(orders.values())

    LOB_bids = {}
    LOB_asks = {}

    for order in order_book:
        order_remaining = (
            round(order.remaining, mm_agent.market_position_decimal)
            if hasattr(mm_agent, "market_position_decimal")
            else order.remaining
        )
        if order.side == vega_protos.vega.Side.SIDE_BUY:
            if order.price not in LOB_bids:
                LOB_bids[order.price] = order_remaining
            else:
                LOB_bids[order.price] += order_remaining
        else:
            if order.price not in LOB_asks:
                LOB_asks[order.price] = order_remaining
            else:
                LOB_asks[order.price] += order_remaining

    return {
        "Order Book Bid Side": LOB_bids,
        "Order Book Ask Side": LOB_asks,
    }


def momentum_trader_data_extraction(
    vega: VegaServiceNull,
    agents: List[Agent],
) -> Dict[str, Any]:
    trader = [agent for agent in agents if isinstance(agent, MomentumTrader)][0]

    general, margin, _ = vega.party_account(
        wallet_name=trader.wallet_name,
        asset_id=trader.asset_id,
        market_id=trader.market_id,
    )

    position = vega.positions_by_market(
        wallet_name=trader.wallet_name, market_id=trader.market_id
    )

    if not position:
        realised_pnl = 0
        unrealised_pnl = 0
        inventory = 0
        entry_price = 0
    else:
        realised_pnl = round(float(position[0].realised_pnl), trader.adp)
        unrealised_pnl = round(float(position[0].unrealised_pnl), trader.adp)
        inventory = float(position[0].open_volume)
        entry_price = float(position[0].average_entry_price) / 10**trader.mdp

    market_data = vega.market_data(market_id=trader.market_id)
    markprice = float(market_data.mark_price) / 10**trader.mdp
    mid_price = float(market_data.mid_price) / 10**trader.mdp
    trading_mode = market_data.market_trading_mode

    market_state = vega.market_info(market_id=trader.market_id).state

    logs = {
        "MT: General Account": general,
        "MT: Margin Account": margin,
        "MT: GeneralPnl": general + margin - trader.initial_asset_mint,
        "MT: RealisedPnl": realised_pnl,
        "MT: UnrealisedPnl": unrealised_pnl,
        "MT: Position": inventory,
        "MT: entry price": entry_price,
        "MT: Indicator": trader.indicators[-1],
        "MidPrice": mid_price,
        "Markprice": markprice,
        "Market State": market_state,
        "Market Trading Mode": trading_mode,
    }
    return logs


def uninformed_tradingbot_data_extraction(
    vega: VegaServiceNull,
    agents: List[Agent],
) -> Dict[str, Any]:
    trader = [agent for agent in agents if isinstance(agent, MarketOrderTrader)][0]

    general, margin, _ = vega.party_account(
        wallet_name=trader.wallet_name,
        asset_id=trader.asset_id,
        market_id=trader.market_id,
    )

    position = vega.positions_by_market(
        wallet_name=trader.wallet_name, market_id=trader.market_id
    )

    if not position:
        realised_pnl = 0
        unrealised_pnl = 0
        inventory = 0
        entry_price = 0
    else:
        realised_pnl = round(float(position[0].realised_pnl), trader.adp)
        unrealised_pnl = round(float(position[0].unrealised_pnl), trader.adp)
        inventory = float(position[0].open_volume)
        entry_price = float(position[0].average_entry_price) / 10**trader.mdp

    logs = {
        "UT: General Account": general,
        "UT: Margin Account": margin,
        "UT: GeneralPnl": general + margin - trader.initial_asset_mint,
        "UT: RealisedPnl": realised_pnl,
        "UT: UnrealisedPnl": unrealised_pnl,
        "UT: Position": inventory,
        "UT: entry price": entry_price,
    }
    return logs



"""market.py

Module contains classes to build a MarketConfig object containing market parameters
which can be passed to service methods or market manager agents to propose markets.

A MarketConfig class has the following attributes which can be set:

• decimal_places
• position_decimal_places
• meta_data
• price_monitoring_parameters.horizon
• price_monitoring_parameters.probability
• price_monitoring_parameters.auction_extension
• liquidity_monitoring_parameters.triggering_ratio
• liquidity_monitoring_parameters.auction_extension
• liquidity_monitoring_parameters.target_stake_parameters.time_window
• liquidity_monitoring_parameters.target_stake_parameters.scaling_factor
• log_normal.tau
• log_normal.risk_aversion_parameters
• log_normal.params.mu
• log_normal.params.r
• log_normal.params.sigma
• instrument.name
• instrument.code
• instrument.future.settlement_asset
• instrument.future.quote_name
• instrument.future.settlement_data_decimals
• instrument.future.terminating_key


Examples:

    A default MarketConfig object can be built with the following:

    $ market_config = MarketConfig("default")


    Pre-configured liquidity monitoring parameters can be loaded with the following:

    $ market_config = market_config.liquidity_monitoring_parameters.load("strict")


    Individual market parameter settings can be set with the following:

    $ market_config = market_config.set("decimal_places", 4)


    The NewMarketConfiguration proto can be built with the following:

    $ new_market_configuration_proto = market_config.build()

"""

import functools
import logging
from typing import Optional

import vega_sim.proto.vega as vega_protos
import vega_sim.proto.vega.data.v1 as oracles_protos
import vega_sim.proto.vega.data_source_pb2 as data_source_protos


def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition(".")
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


def rgetattr(obj, attr, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return functools.reduce(_getattr, [obj] + attr.split("."))


class Config:
    OPTS = {}

    def __init__(self, opt: Optional[str] = None) -> None:
        self.load(opt=opt)

    def load(self, opt: Optional[str] = None):
        if opt is None:
            opt = list(self.OPTS.keys())[0]
            logging.debug(f"No 'opt' arg given. Using default value '{opt}'.")

        if opt not in self.OPTS:
            raise ValueError(f"Invalid 'opt' arg '{opt}' specified.")

        return opt


class MarketConfig(Config):
    OPTS = {
        "default": {
            "decimal_places": 4,
            "position_decimal_places": 2,
            "metadata": None,
            "price_monitoring_parameters": "default",
            "liquidity_monitoring_parameters": "default",
            "log_normal": "default",
            "instrument": "default",
        }
    }

    def load(self, opt: Optional[str] = None):
        opt = super().load(opt=opt)

        self.decimal_places = self.OPTS[opt]["decimal_places"]
        self.position_decimal_places = self.OPTS[opt]["position_decimal_places"]
        self.metadata = self.OPTS[opt]["metadata"]

        self.instrument = InstrumentConfiguration(opt=self.OPTS[opt]["instrument"])
        self.price_monitoring_parameters = PriceMonitoringParameters(
            opt=self.OPTS[opt]["price_monitoring_parameters"]
        )
        self.liquidity_monitoring_parameters = LiquidityMonitoringParameters(
            opt=self.OPTS[opt]["liquidity_monitoring_parameters"]
        )
        self.log_normal = LogNormalRiskModel(opt=self.OPTS[opt]["log_normal"])

    def build(self):
        return vega_protos.governance.NewMarket(
            changes=vega_protos.governance.NewMarketConfiguration(
                decimal_places=self.decimal_places,
                position_decimal_places=self.position_decimal_places,
                metadata=self.metadata,
                instrument=self.instrument.build(),
                price_monitoring_parameters=self.price_monitoring_parameters.build(),
                liquidity_monitoring_parameters=self.liquidity_monitoring_parameters.build(),
                log_normal=self.log_normal.build(),
            )
        )

    def set(self, parameter, value):
        rsetattr(self, attr=parameter, val=value)


class PriceMonitoringParameters(Config):
    OPTS = {
        "default": {
            "horizon": 24 * 3600,
            "probability": "0.999999",
            "auction_extension": 5,
        }
    }

    def load(self, opt: Optional[str] = None):
        opt = super().load(opt=opt)

        self.horizon = self.OPTS[opt]["horizon"]
        self.probability = self.OPTS[opt]["probability"]
        self.auction_extension = self.OPTS[opt]["auction_extension"]

    def build(self):
        return vega_protos.markets.PriceMonitoringParameters(
            triggers=[
                {
                    "horizon": self.horizon,
                    "probability": self.probability,
                    "auction_extension": self.auction_extension,
                }
            ]
        )


class LiquidityMonitoringParameters(Config):
    OPTS = {
        "default": {
            "triggering_ratio": 0.7,
            "auction_extension": 0,
            "target_stake_parameters": "default",
        }
    }

    def load(self, opt: Optional[str] = None):
        opt = super().load(opt=opt)

        self.triggering_ratio = self.OPTS[opt]["triggering_ratio"]
        self.auction_extension = self.OPTS[opt]["auction_extension"]

        self.target_stake_parameters = TargetStakeParameters(
            opt=self.OPTS[opt]["target_stake_parameters"]
        )

    def build(self):
        return vega_protos.markets.LiquidityMonitoringParameters(
            triggering_ratio=self.triggering_ratio,
            auction_extension=self.auction_extension,
            target_stake_parameters=self.target_stake_parameters.build(),
        )


class TargetStakeParameters(Config):
    OPTS = {
        "default": {
            "time_window": 60 * 60,
            "scaling_factor": 1,
        }
    }

    def load(self, opt: Optional[str] = None):
        opt = super().load(opt=opt)

        self.time_window = self.OPTS[opt]["time_window"]
        self.scaling_factor = self.OPTS[opt]["scaling_factor"]

    def build(self):
        return vega_protos.markets.TargetStakeParameters(
            time_window=self.time_window,
            scaling_factor=self.scaling_factor,
        )


class LogNormalRiskModel(Config):
    OPTS = {
        "default": {
            "risk_aversion_parameter": 0.01,
            "tau": 1.90128526884173e-06,
            "params": "default",
        }
    }

    def load(self, opt: Optional[str] = None):
        opt = super().load(opt=opt)

        self.risk_aversion_parameter = self.OPTS[opt]["risk_aversion_parameter"]
        self.tau = self.OPTS[opt]["tau"]

        self.params = LogNormalModelParams(opt=self.OPTS[opt]["params"])

    def build(self):
        return vega_protos.markets.LogNormalRiskModel(
            risk_aversion_parameter=self.risk_aversion_parameter,
            tau=self.tau,
            params=self.params.build(),
        )


class LogNormalModelParams(Config):
    OPTS = {
        "default": {
            "mu": 0,
            "r": 0.016,
            "sigma": 3.0,
        }
    }

    def __init__(self, opt: Optional[str] = None) -> None:
        super().__init__(opt)

    def load(self, opt: Optional[str] = None):
        opt = super().load(opt=opt)

        self.mu = self.OPTS[opt]["mu"]
        self.r = self.OPTS[opt]["r"]
        self.sigma = self.OPTS[opt]["sigma"]

    def build(self):
        return vega_protos.markets.LogNormalModelParams(
            mu=self.mu,
            r=self.r,
            sigma=self.sigma,
        )


class InstrumentConfiguration(Config):
    OPTS = {
        "default": {
            "name": None,
            "code": None,
            "future": "default",
        }
    }

    def load(self, opt: Optional[str] = None):
        opt = super().load(opt=opt)

        self.name = self.OPTS[opt]["name"]
        self.code = self.OPTS[opt]["code"]
        self.future = FutureProduct(opt=self.OPTS[opt]["future"])

    def build(self):
        return vega_protos.governance.InstrumentConfiguration(
            name=self.name, code=self.code, future=self.future.build()
        )


class FutureProduct(Config):
    OPTS = {
        "default": {
            "settlement_asset": None,
            "quote_name": None,
            "settlement_data_decimals": None,
            "terminating_key": None,
        }
    }

    def load(self, opt: Optional[str] = None):
        opt = super().load(opt=opt)

        self.settlement_asset = self.OPTS[opt]["settlement_asset"]
        self.quote_name = self.OPTS[opt]["quote_name"]
        self.settlement_data_decimals = self.OPTS[opt]["settlement_data_decimals"]
        self.terminating_key = self.OPTS[opt]["terminating_key"]

    def build(self):
        if None in [
            self.settlement_asset,
            self.quote_name,
            self.settlement_data_decimals,
            self.terminating_key,
        ]:
            print(
                [
                    self.settlement_asset,
                    self.quote_name,
                    self.settlement_data_decimals,
                    self.terminating_key,
                ]
            )
            raise ValueError(
                f"MarketConfig has not been updated with settlement asset information."
            )

        data_source_spec_for_settlement_data = data_source_protos.DataSourceDefinition(
            external=data_source_protos.DataSourceDefinitionExternal(
                oracle=data_source_protos.DataSourceSpecConfiguration(
                    signers=[
                        oracles_protos.data.Signer(
                            pub_key=oracles_protos.data.PubKey(key=self.terminating_key)
                        )
                    ],
                    filters=[
                        oracles_protos.spec.Filter(
                            key=oracles_protos.spec.PropertyKey(
                                name=f"price.{self.quote_name}.value",
                                type=oracles_protos.spec.PropertyKey.Type.TYPE_INTEGER,
                            ),
                            conditions=[],
                        )
                    ],
                )
            )
        )
        data_source_spec_for_trading_termination = data_source_protos.DataSourceDefinition(
            external=data_source_protos.DataSourceDefinitionExternal(
                oracle=data_source_protos.DataSourceSpecConfiguration(
                    signers=[
                        oracles_protos.data.Signer(
                            pub_key=oracles_protos.data.PubKey(key=self.terminating_key)
                        )
                    ],
                    filters=[
                        oracles_protos.spec.Filter(
                            key=oracles_protos.spec.PropertyKey(
                                name="trading.terminated",
                                type=oracles_protos.spec.PropertyKey.Type.TYPE_BOOLEAN,
                            ),
                            conditions=[],
                        )
                    ],
                )
            )
        )
        data_source_spec_binding = vega_protos.markets.DataSourceSpecToFutureBinding(
            settlement_data_property=f"price.{self.quote_name}.value",
            trading_termination_property="trading.terminated",
        )

        return vega_protos.governance.FutureProduct(
            settlement_asset=self.settlement_asset,
            quote_name=self.quote_name,
            data_source_spec_for_settlement_data=data_source_spec_for_settlement_data,
            data_source_spec_for_trading_termination=data_source_spec_for_trading_termination,
            data_source_spec_binding=data_source_spec_binding,
            settlement_data_decimals=self.settlement_data_decimals,
        )



import torch
import torch.nn as nn
from typing import List


class Softmax(nn.Softmax):
    def __init__(self):
        super().__init__()
        self.activation = nn.Softmax(dim=-1)

    def forward(self, x):
        return self.activation(x)


class Threshold(nn.Module):
    def __init__(self, threshold):
        super().__init__()
        self.threshold = threshold

    def forward(self, x):
        return nn.LogSigmoid()(x) + self.threshold


class FFN(nn.Module):
    def __init__(
        self, sizes, activation=nn.ReLU, output_activation=nn.Identity, batch_norm=False
    ):
        super().__init__()

        layers = []
        for j in range(1, len(sizes)):
            layers.append(nn.Linear(sizes[j - 1], sizes[j]))
            if j < (len(sizes) - 1):
                layers.append(activation())
            else:
                layers.append(output_activation())

        self.net = nn.Sequential(*layers)

    def forward(self, *args):
        x = torch.cat(args, -1)
        return self.net(x)


class FFN_fix_fol_Q(nn.Module):
    """
    Specific class for Q-func of trader
    """

    def __init__(
        self,
        state_dim: int,
    ):
        """
        I create as many sub-networks as possible discrete actions that the agent can take.
        The input of each network is the continuous part of the action (i.e. the volume).
        The output of each network is the q-func applied to the state, the volume and that particular discrete action
        """
        super().__init__()

        self.ffn_sell = FFN(
            sizes=[state_dim, 4096, 1], activation=nn.ReLU
        )  # +action_cont_dim is for the volume
        self.ffn_buy = FFN(
            sizes=[state_dim, 4096, 1], activation=nn.ReLU
        )  # +action_cont_dim is for the volume
        self.ffn_do_nothing = FFN(sizes=[state_dim, 4096, 1], activation=nn.ReLU)

    def forward(self, state: torch.Tensor):
        output = []
        output.append(self.ffn_sell(state))
        output.append(self.ffn_buy(state))
        output.append(self.ffn_do_nothing(state))
        return torch.cat(output, -1)


class FFN_Q(nn.Module):
    """
    Specific class for Q-func of trader
    """

    def __init__(
        self,
        state_dim: int,
    ):
        """
        I create as many sub-networks as possible discrete actions that the agent can take.
        The input of each network is the continuous part of the action (i.e. the volume).
        The output of each network is the q-func applied to the state, the volume and that particular discrete action
        """
        super().__init__()

        self.ffn_sell = FFN(
            sizes=[state_dim + 1, 32, 1], activation=nn.Tanh
        )  # +action_cont_dim is for the volume
        self.ffn_buy = FFN(
            sizes=[state_dim + 1, 32, 1], activation=nn.Tanh
        )  # +action_cont_dim is for the volume
        self.ffn_do_nothing = FFN(sizes=[state_dim, 32, 1], activation=nn.Tanh)

    def forward(
        self, state: torch.Tensor, volume_sell: torch.Tensor, volume_buy: torch.Tensor
    ):
        if volume_sell.dim() == 1:
            volume_sell = volume_sell.unsqueeze(1)
        if volume_buy.dim() == 1:
            volume_buy = volume_buy.unsqueeze(1)
        output = []
        output.append(self.ffn_sell(state, volume_sell))
        output.append(self.ffn_buy(state, volume_buy))
        output.append(self.ffn_do_nothing(state))
        return torch.cat(output, -1)


class FFN_Params_Normal(nn.Module):
    def __init__(
        self,
        n_in: int,
        n_distr: int,
        hidden_sizes: List[int],
    ):

        super().__init__()
        self.net = FFN(
            sizes=[n_in] + hidden_sizes, activation=nn.Tanh, output_activation=nn.Tanh
        )
        self.net_mu = nn.Linear(hidden_sizes[-1], n_distr)
        self.net_sigma = nn.Sequential(
            nn.Linear(hidden_sizes[-1], n_distr), nn.Softplus()
        )

    def forward(self, *args):
        x = torch.cat(args, -1)
        y = self.net(x)
        mu = self.net_mu(y)
        sigma = self.net_sigma(y)
        return mu, sigma


@torch.no_grad()
def init_weights(m):
    if isinstance(m, nn.Linear):
        m.weight.fill_(0.0)
        m.bias.fill_(0.0)



from importlib import metadata
from pathlib import Path


__version__ = metadata.version(__package__)

vega_home_path = Path(__file__).parent / "vegahome"
vega_bin_path = Path(__file__).parent / "bin"



import numpy as np
from scipy.linalg import expm


def A_S_MMmodel(
    T: float,
    dt: float,
    length: int,
    mdp: int,
    q_upper: int,
    q_lower: int,
    kappa: int,
    lmbda: int,
    alpha: float,
    phi: float,
):

    """
    Simulate an Market Making Model and output the optimal strategy
      of posting bid/ask depth based on market parameters.

    Args:
        T:
            int, total simulation time
        dt:
            float, the increment of each time step
        length:
            int, total time steps
        mdp:
            int, market decimal place
        q_upper:
            int, upper bound of the inventory of MM specified in the model
        q_lower:
            int, lower bound of the inventory of MM specified in the model
        kappa:
            int, market parameter to represnet the probability of pegged LOs to be hit
        lmbda:
            int, market parameter to represnet the coming rate of MOs
        alpha:
            float, risk aversion parameter to represnet terminal penalty coefficient
        phi:
            float, risk aversion parameter to represnet running penalty coefficient
    """
    # Unify unit to minute in MM Model
    T *= 60 * 24 * 365.25
    dt *= 60 * 24 * 365.25

    # Initialize 2-d arrays to store optimal strategy
    optimal_depth_bid = np.zeros([length, q_upper - q_lower])
    optimal_depth_ask = np.zeros([length, q_upper - q_lower])

    # Let A be a (q_upper-q_lower+1)-square matrix
    A = np.zeros([q_upper - q_lower + 1, q_upper - q_lower + 1])
    # w, time * (q_upper-q_lower+1)-dim matrix, to store the solution of ODE
    #   row corresponds to time_step, column corresponds to inventory q
    w = np.zeros([length, q_upper - q_lower + 1])

    # A is the coefficient matrix of ODE
    for i in range(q_upper - q_lower + 1):
        for j in range(q_upper - q_lower + 1):
            # i denotes row/ j denotes column
            if j == i:
                A[i, j] = -kappa * phi * (q_upper - i) ** 2
            elif j == i + 1:
                A[i, j] = lmbda * np.e**-1
            elif j == i - 1:
                A[i, j] = lmbda * np.e**-1

    # z, (q_upper-q_lower+1)-dim vector, denotes the terminal condition of ODE
    z = np.array(
        [np.exp(-alpha * kappa * j**2) for j in range(q_upper, q_lower - 1, -1)]
    )

    for i in range(length):
        # at each time_step
        w[i, :] = np.dot(expm(A * (T - i * dt)), z)

    # h is the transformation of solution from ODE
    #   also the key term of value function
    h = np.log(w) / kappa

    # Calculate optimal strategy
    for i in range(q_upper - q_lower):
        # column corresponds to Q, Q-1,..., -Q+1
        optimal_depth_ask[:, i] = 1 / kappa + h[:, i] - h[:, i + 1]

    for i in range(1, q_upper - q_lower + 1):
        # column corresponds to Q-1, Q-2,..., -Q
        optimal_depth_bid[:, i - 1] = 1 / kappa + h[:, i] - h[:, i - 1]

    # In A_S model, optimal depth can be negative, however, in Vega,
    #   offset must be positive and notice the market precision
    optimal_depth_bid = np.round(optimal_depth_bid, mdp)
    optimal_depth_ask = np.round(optimal_depth_ask, mdp)
    # So, if the depth is negative, replace with the minimum tick size
    optimal_depth_bid[optimal_depth_bid <= 0] = 1 / 10**mdp
    optimal_depth_ask[optimal_depth_ask <= 0] = 1 / 10**mdp

    return optimal_depth_bid, optimal_depth_ask, h


def GLFT_approx(
    q_upper: int,
    q_lower: int,
    kappa: int,
    lmbda: int,
    alpha: float,
    phi: float,
):
    """
    One problem of A_S Market Making model is cannot handle large exponential matrix
      when terminal time T, or market parameters kappa/lmbda, are large enough.
      The GLFT formula is to explore the asymptotic behavior of the optimal strategy,
      assuming T -> infty.

     Args:
        q_upper:
            int, upper bound of the inventory of MM specified in the model
        q_lower:
            int, lower bound of the inventory of MM specified in the model
        kappa:
            int, market parameter to represnet the probability of pegged LOs to be hit
        lmbda:
            int, market parameter to represnet the coming rate of MOs
        alpha:
            float, risk aversion parameter to represnet terminal penalty coefficient
        phi:
            float, risk aversion parameter to represnet running penalty coefficient
    """

    # approximation method
    delta_buy_approx = [
        1 / kappa + (2 * i + 1) * np.sqrt(phi * np.e / lmbda / kappa) / 2
        for i in range(q_upper - 1, q_lower - 1, -1)
    ]
    delta_sell_approx = [
        1 / kappa - (2 * i - 1) * np.sqrt(phi * np.e / lmbda / kappa) / 2
        for i in range(q_upper, q_lower, -1)
    ]

    return delta_buy_approx, delta_sell_approx
